#include<iostream>

// Small "boundary" matrix (sbmatrix) class (everything inlined)
// Construct an (Ndim x Ndim-1) matrix used for parameterizing 
// surfaces.

template<const int Ndim> class svector;
template<const int Ndim> class smatrix;
template<const int Ndim> class sbmatrix;
template<const int Ndim>
           sbmatrix<Ndim> operator*(double c, const sbmatrix<Ndim>& a);

template<const int Ndim>
           sbmatrix<Ndim> operator*(const smatrix<Ndim>& a, 
				    const sbmatrix<Ndim>& b);

template<const int Ndim>
class sbmatrix 
{
public:
  sbmatrix();
  sbmatrix(const sbmatrix&);
  sbmatrix(const double a[Ndim][Ndim-1]);
  sbmatrix(const svector<Ndim>&, const svector<Ndim-1>&);  // outer product

  sbmatrix& zero();                // reset to zero

  sbmatrix operator+(const sbmatrix&) const;
  sbmatrix& operator+=(const sbmatrix&);
  sbmatrix operator-() const;
  sbmatrix operator-(const sbmatrix&) const;
  svector<Ndim> operator*(const svector<Ndim-1>&) const;

  double frobNorm() const;

  const svector<Ndim> operator[](int i) const;
  svector<Ndim>& operator[](int i); 

  svector<Ndim> normal(double &det) const;

  // private:
  svector<Ndim> m[Ndim-1];   // m[j] = i-th column of the matrix

  // friend sbmatrix<Ndim> operator*<Ndim>(double c, const sbmatrix<Ndim> &a);
  // friend sbmatrix<Ndim> operator*<Ndim>(const smatrix<Ndim> &a, 
  //		  			   const sbmatrix<Ndim> &b);
};


template<const int Ndim>
inline double sbmatrix<Ndim>::frobNorm() const
{
  double norm = 0.0;

  for (int j = 0; j < Ndim-1; j++) norm += m[j].dot(m[j]);

  return sqrt(norm);  // compute the froebenius norm of this matrix
}

template<const int Ndim>
inline sbmatrix<Ndim>::sbmatrix()
{
  for(int j = 0; j < Ndim-1; j++) m[j].zero();
}

template<const int Ndim>
inline sbmatrix<Ndim>::sbmatrix(const sbmatrix<Ndim>& a)
{
  for(int j = 0; j < Ndim-1; j++) m[j] = a.m[j];
 }

template<const int Ndim>
inline sbmatrix<Ndim>::sbmatrix(const double a[Ndim][Ndim-1])
{
  for(int i = 0; i < Ndim; i++)
  for(int j = 0; j < Ndim-1; j++) m[j][i] = a[i][j];
}

template<const int Ndim>
inline sbmatrix<Ndim>::sbmatrix(const svector<Ndim>& v, 
				const svector<Ndim-1>& w)
{
  for(int i = 0; i < Ndim; i++)
  for(int j = 0; j < Ndim-1; j++) m[j][i] = v[i] * w[j];
}

template<const int Ndim>
inline sbmatrix<Ndim>& sbmatrix<Ndim>::zero()
{
  for(int j = 0; j < Ndim-1; j++) m[j].zero();

  return(*this);
}

template<const int Ndim>
inline sbmatrix<Ndim> sbmatrix<Ndim>::operator+(const sbmatrix<Ndim>& a) const
{
  sbmatrix<Ndim> b;

  for(int j = 0; j < Ndim-1; j++) b.m[j] = m[j] + a.m[j];

  return(b);
}

template<const int Ndim>
inline sbmatrix<Ndim>& sbmatrix<Ndim>::operator+=(const sbmatrix<Ndim>& a)
{
  for(int j = 0; j < Ndim-1; j++) m[j] += a.m[j];

  return(*this);
}

template<const int Ndim>
inline sbmatrix<Ndim> sbmatrix<Ndim>::operator-() const
{
  sbmatrix<Ndim> b;

  for(int j = 0; j < Ndim-1; j++) b.m[j] = -m[j];

  return(b);
}

template<const int Ndim>
inline sbmatrix<Ndim> sbmatrix<Ndim>::operator-(const sbmatrix<Ndim>& a) const
{
  sbmatrix<Ndim> b;

  for(int j = 0; j < Ndim-1; j++) b.m[j] = m[j] - a.m[j];

  return(b);
}

template<const int Ndim>
inline sbmatrix<Ndim> operator*(double c, const sbmatrix<Ndim>& a)
{
  sbmatrix<Ndim> b;

  for(int j = 0; j < Ndim-1; j++) b.m[j] = c * a.m[j];

  return(b);
}

template<const int Ndim>
inline sbmatrix<Ndim> operator*(const smatrix<Ndim>& a,
				const sbmatrix<Ndim>& b)
{
  sbmatrix<Ndim> c;

  for(int i = 0; i < Ndim; i++)
  for(int j = 0; j < Ndim-1; j++)
    {
      c[j][i] = 0.0;

      for(int k = 0; k < Ndim; k++) c[j][i] += a[i][k]*b[j][k];
    }

  return(c);
}

template<const int Ndim>
inline svector<Ndim> sbmatrix<Ndim>::operator*(const svector<Ndim-1>& v) const
{
  svector<Ndim> w;

  for(int i = 0; i < Ndim; i++) 
    {
      w.x[i] = 0.0;

      for(int j = 0; j < Ndim-1; j++) w.x[i] += m[j][i] * v.x[j];
    }

  return(w);
}


template<const int Ndim>
inline svector<Ndim>& sbmatrix<Ndim>::operator[](int j)
{
  return(m[j]);
}

template<const int Ndim>
inline const svector<Ndim> sbmatrix<Ndim>::operator[](int j) const
{
  // Note: Returning "const svector<Ndim>" guarantees that the
  //  syntax "a[i][j] = c" will not just update a copy of m[i]

  return(m[j]);
}

template<const int Ndim>
svector<Ndim> sbmatrix<Ndim>::normal(double &det) const
{
  if(Ndim == 2)
    {
      svector<Ndim> nn(m[0][0], m[0][1]);
      
      det = nn.norm();

      assert(det != 0.0);

      return( svector<Ndim>(nn[1]/det, -nn[0]/det) );
    }
  else if(Ndim == 3)
    {
      svector<Ndim> nn(m[0][1]*m[1][2]-m[0][2]*m[1][1],
		    m[0][2]*m[1][0]-m[0][0]*m[1][2],
		    m[0][0]*m[1][1]-m[0][1]*m[1][0]);  // cross product

      det = nn.norm();

      assert(det != 0.0);

      return( (1/det)*nn );
    }
  else
    {
      throw("sbmatrix<Ndim>::normal() only implemented for Ndim = 2 and 3");
    }
}
    
