// Set up a mesh with each simplex affine equivalent to the unit tetrahedra

#include<string>
#include<fstream>
#include<map>
#include<vector>
#include"../foreach.hpp"

#include "gelement3d.C"

using namespace std;

class dir3Mesh
{
public:
    dir3Mesh(string);                        // read mesh.poly file
    dir3Mesh(const dir3Mesh &mm) 
    {
	cerr << "dir3Mesh(): Don't copy me, deep copy nyi" << endl;
	throw("don't copy a dir3Mesh");
     }
    dir3Mesh() 
    {
	cerr << "dir3Mesh(): Default constructor called, nyi" << endl;
	throw("bad dir3Mesh constructor");
    }
    
    ~dir3Mesh() { return; }
   
    int nelem;
    int nodes;
    int nbelem;
    
    void print(ostream& dump);
   
    class geiterator {
    public:
	geiterator();
	geiterator(int l, dir3Mesh * g);
	void operator++();
	void operator++(int);
	geiterator& operator=(const geiterator & it);
	bool operator==(const geiterator & it);
	bool operator!=(const geiterator & it);
        gelement operator*();

    private:
	int loc;
	dir3Mesh * gm;
    };

    geiterator gebegin();
    geiterator geend();

    class bgeiterator {
    public:
	bgeiterator();
	bgeiterator(int l, dir3Mesh * g);
	void operator++();
	void operator++(int);
	bgeiterator& operator=(const bgeiterator & it);
	bool operator==(const bgeiterator & it);
	bool operator!=(const bgeiterator & it);
        bgelement operator*();

    private:
	int loc;
	dir3Mesh * gm;
    };

    bgeiterator bgebegin();
    bgeiterator bgeend();

private:

  map<int, svector<3> > vertices;    // vertices[id] = vertex coordinates

  vector<intArray<3> > triangles;   // boundary triangles
  vector<int> triId;                // triId[0..nbelem-1] = triId

  vector<intArray<4> > tets;        // tetrahedra

  void input(string meshFile);       // auxillary function to read .tplc file
};

dir3Mesh::dir3Mesh(string meshFile) 
{
  cout << "Initializing dir3Mesh(" << meshFile << "):" 
       << " (reads .tplc file)" << endl;
  
  input(meshFile);
  
  cout << "Number of vertices read in = ..................... " << nodes
       << endl;
  cout << "Number of boundary triangles = ................... " << nbelem 
       << endl;
  cout << "Number of tetrahedra = ........................... " << nelem 
       << endl;
  
  return;
}

dir3Mesh::geiterator::geiterator() {
    loc = 0;
    gm = NULL;
}

dir3Mesh::geiterator::geiterator(int l, dir3Mesh * g) {
    loc = l;
    gm = g;
}


void dir3Mesh::geiterator::operator++() {
    (*this)++;
}

void dir3Mesh::geiterator::operator++(int) {
    loc++;
}

dir3Mesh::geiterator& dir3Mesh::geiterator::operator=(const geiterator & it) {
    loc = it.loc;
    gm = it.gm;
    return *this;
}

bool dir3Mesh::geiterator::operator==(const geiterator & it) {
    if (gm != it.gm || loc != it.loc) {
	return false;
    }
    return true;
}

bool dir3Mesh::geiterator::operator!=(const geiterator & it) {
    return !(*this == it);
}

gelement dir3Mesh::geiterator::operator*() {

    gelement ee;
    int j, *tt = gm->tets[loc]; 

    ee.master = &LinearTetrahedron;
    
    for(j = 0; j < 4; j++) 
    {
	ee.xx[j] = gm->vertices[tt[j]];
	ee.nodes[j] = tt[j];
    }
    
    return ee;
}


dir3Mesh::geiterator dir3Mesh::gebegin() {
    return geiterator(0, this);
}

dir3Mesh::geiterator dir3Mesh::geend() {
    return geiterator(nelem, this);
}

dir3Mesh::bgeiterator::bgeiterator() {
    loc = 0;
    gm = NULL;
}

dir3Mesh::bgeiterator::bgeiterator(int l, dir3Mesh * g) {
    loc = l;
    gm = g;
}

void dir3Mesh::bgeiterator::operator++() {
    (*this)++;
}

void dir3Mesh::bgeiterator::operator++(int) {
    loc++;
}

dir3Mesh::bgeiterator& dir3Mesh::bgeiterator::operator=(const bgeiterator & it) {
    loc = it.loc;
    gm = it.gm;
    return *this;
}

bool dir3Mesh::bgeiterator::operator==(const bgeiterator & it) {
    if (loc != it.loc || gm != it.gm) {
	return false;
    }
    return true;
}

bool dir3Mesh::bgeiterator::operator!=(const bgeiterator & it) {
    return !(*this == it);
}

bgelement dir3Mesh::bgeiterator::operator*() {

  int j, *tt = gm->triangles[loc];
  bgelement be;

  be.bctype = gm->triId[loc];

  be.master = &LinearTriangle;

  for(j = 0; j < 3; j++) 
  {
      be.xx[j] = gm->vertices[tt[j]];
      
      be.nodes[j] = tt[j];
  }  
  return be;

}

dir3Mesh::bgeiterator dir3Mesh::bgebegin() {
    return bgeiterator(0,this);
}

dir3Mesh::bgeiterator dir3Mesh::bgeend() {
    return bgeiterator(nbelem,this);
}

void dir3Mesh::input(string meshFile)
{
  int i, j, dim, id, i0,i1, i2, i3, nFace, nSegment, np, ntri, faceId;
  double xx, yy, zz;

  ifstream in((meshFile).c_str());

  if( !in.is_open() ) in.open((meshFile + ".tplc").c_str());

  if( !in.is_open() ) 
    {
      cerr << "dir3Mesh::input(): can't open .tplc file:" << meshFile << endl;

      throw("Can't open file");
    }

  in.ignore(256,'\n');    // first line is TPLC
  in.ignore(256,' ');     // Second line is DIMENSION dim
  in >> dim;
  assert(dim == 3);

  in.ignore(256,'\n');  
  in.ignore(256,' ');    // Next line is POINTS nodes
  in >> nodes;
  assert(nodes > 0);

  for(i = 0; i < nodes; i++)
    {
      in.ignore(256,'\n');

      in >> id;

      in >> xx >> yy >> zz;

      vertices[id] = svector<3>(xx,yy,zz);
    }

  // Read and discard segments

  in.ignore(256,'\n');
  in.ignore(256,' ');       // next line is SEGMENTS nSegment
  in >> nSegment;
  assert(nSegment >= 0);
  
  for(i = 0; i < nSegment; i++)
    {
    in.ignore(256,'\n');
   
    in >> id >> np;

    for(j = 0; j < np; j++) in >> i0; 
    }

  // Read the triangulated faces

  in.ignore(256,'\n');
  in.ignore(256,' ');       // next line is FACES nFace
  in >> nFace;
  assert(nFace >= 0);

  nbelem = 0;

  for(i = 0; i < nFace; i++)
    {
    in.ignore(256,'\n');

    in >> faceId >> ntri;

    assert(ntri >= 0);

    triId.resize(nbelem + ntri);
    triangles.resize(nbelem + ntri);

    for(j = 0; j < ntri; j++)
      {
      in.ignore(256,'\n');

      in >> i0 >> i1 >> i2;

      assert(vertices.find(i0) != vertices.end());
      assert(vertices.find(i1) != vertices.end());
      assert(vertices.find(i2) != vertices.end());

      triangles[nbelem][0] = i0; 
      triangles[nbelem][1] = i1; 
      triangles[nbelem][2] = i2; 

      triId[nbelem++] = faceId;
      }
    }

  // Read the tetrahedra

  in.ignore(256,'\n');
  in.ignore(256,' ');       // next line is TETS nTet
  in >> nelem;
  assert(nelem >= 0);

  tets.resize(nelem);

  for(i = 0; i < nelem; i++)
    {
    in.ignore(256,'\n');

    in >> i0 >> i1 >> i2 >> i3;

    assert(vertices.find(i0) != vertices.end());
    assert(vertices.find(i1) != vertices.end());
    assert(vertices.find(i2) != vertices.end());
    assert(vertices.find(i3) != vertices.end());

    tets[i][0] = i0; 
    tets[i][1] = i1; 
    tets[i][2] = i2; 
    tets[i][3] = i3; 
    }

  in.close();

  return;
}
